"""
Cart pole swing-up: Original version from:
https://github.com/zuoxingdong/DeepPILCO/blob/master/cartpole_swingup.py

Modified so that done=True when x is outside of -2.4 to 2.4
Reward is also reshaped to be similar to PyBullet/roboschool version

More difficult, since dt is 0.05 (not 0.01), and only 200 timesteps
"""

import logging
import math
import gym
from gym import spaces
from gym.utils import seeding
import numpy as np

logger = logging.getLogger(__name__)

class CartPoleSwingUpEnv(gym.Env):
    metadata = {
        'render.modes': ['human', 'rgb_array'],
        'video.frames_per_second' : 50
    }

    def __init__(self):
        self.g = 9.82  # gravity
        self.m_c = 0.5  # cart mass
        self.m_p = 0.5  # pendulum mass
        self.total_m = (self.m_p + self.m_c)
        self.l_base = 0.6 # base pole length
        self.l = self.l_base # simulated pole length (see setEnv below)
        self.m_p_l = (self.m_p*self.l)
        self.force_mag = 10.0
        #self.dt = 0.05  # slower reaction (hard mode)
        self.dt = 0.01  # faster reaction (easy mode)
        self.b = 0.1  # friction coefficient

        self.t = 0 # timestep
        #self.t_limit = 200
        self.t_limit = 1000

        # Angle at which to fail the episode
        self.theta_threshold_radians = 12 * 2 * math.pi / 360
        self.x_threshold = 2.4

        high = np.array([
            np.finfo(np.float32).max,
            np.finfo(np.float32).max,
            np.finfo(np.float32).max,
            np.finfo(np.float32).max,
            np.finfo(np.float32).max])

        self.action_space = spaces.Box(-1.0, 1.0, shape=(1,), dtype=float)
        self.observation_space = spaces.Box(-high, high, dtype=float)

        self.seed()
        self.viewer = None
        self.state = None

        self.noise = 0

    def setEnv(self, envChange):
        '''
        Changes the environment, envChange is the percent change of parameter
        '''
        self.l = self.l_base*envChange

    def setNoise(self, noiseVariance):
        '''
        Changes the leven of input noise
        '''
        self.noise = noiseVariance

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def stateUpdate(self,action,state, noise=0):
        x, x_dot, theta, theta_dot = state
        x     += np.random.randn() * noise
        theta += np.random.randn() * noise

        s = math.sin(theta)
        c = math.cos(theta)

        xdot_update = (-2*self.m_p_l*(theta_dot**2)*s + 3*self.m_p*self.g*s*c + 4*action - 4*self.b*x_dot)/(4*self.total_m - 3*self.m_p*c**2)

        thetadot_update = (-3*self.m_p_l*(theta_dot**2)*s*c + 6*self.total_m*self.g*s + 6*(action - self.b*x_dot)*c)/(4*self.l*self.total_m - 3*self.m_p_l*c**2)

        x = x + x_dot*self.dt
        theta = theta + theta_dot*self.dt

        x_dot = x_dot + xdot_update*self.dt
        theta_dot = theta_dot + thetadot_update*self.dt  

        return (x, x_dot, theta, theta_dot)

    def step(self, action):
        # Valid action
        action = np.clip(action, -1.0, 1.0)[0]
        action *= self.force_mag

        noise_obs = self.stateUpdate(action, self.state, noise=self.noise)
        self.state = self.stateUpdate(action, self.state)

        x,x_dot,theta,theta_dot = self.state


        done = False
        if  x < -self.x_threshold or x > self.x_threshold:
          done = True

        self.t += 1
        if self.t >= self.t_limit:
          done = True

        # Reward staying in the middle
        reward_theta = (np.cos(theta)+1.0)/2.0
        reward_x = np.cos((x/self.x_threshold)*(np.pi/2.0))

        reward = reward_theta*reward_x
        #reward = (np.cos(theta)+1.0)/2.0

        x,x_dot,theta,theta_dot = noise_obs
        obs = np.array([x,x_dot,np.cos(theta),np.sin(theta),theta_dot])

        return obs, reward, done, {}

    def reset(self):
        #self.state = self.np_random.normal(loc=np.array([0.0, 0.0, 30*(2*np.pi)/360, 0.0]), scale=np.array([0.0, 0.0, 0.0, 0.0]))
        self.state = np.random.normal(loc=np.array([0.0, 0.0, np.pi, 0.0]), scale=np.array([0.2, 0.2, 0.2, 0.2]))
        self.steps_beyond_done = None
        self.t = 0 # timestep
        x, x_dot, theta, theta_dot = self.state
        obs = np.array([x,x_dot,np.cos(theta),np.sin(theta),theta_dot])
        return obs

    def render(self, mode='human', close=False):
        if close:
            if self.viewer is not None:
                self.viewer.close()
                self.viewer = None
            return

        screen_width = 600
        screen_height = 600 # before was 400

        world_width = 5  # max visible position of cart
        scale = screen_width/world_width
        carty = screen_height/2 # TOP OF CART
        polewidth = 6.0
        polelen = scale*self.l  # 0.6 or self.l
        cartwidth = 40.0
        cartheight = 20.0

        if self.viewer is None:
            from gym.envs.classic_control import rendering
            self.viewer = rendering.Viewer(screen_width, screen_height)
  
            l,r,t,b = -cartwidth/2, cartwidth/2, cartheight/2, -cartheight/2
  
            cart = rendering.FilledPolygon([(l,b), (l,t), (r,t), (r,b)])
            self.carttrans = rendering.Transform()
            cart.add_attr(self.carttrans)
            cart.set_color(1, 0, 0)
            self.viewer.add_geom(cart)
  
            l,r,t,b = -polewidth/2,polewidth/2,polelen-polewidth/2,-polewidth/2
            pole = rendering.FilledPolygon([(l,b), (l,t), (r,t), (r,b)])
            pole.set_color(0, 0, 1)
            self.poletrans = rendering.Transform(translation=(0, 0))
            pole.add_attr(self.poletrans)
            pole.add_attr(self.carttrans)
            self.viewer.add_geom(pole)
  
            self.axle = rendering.make_circle(polewidth/2)
            self.axle.add_attr(self.poletrans)
            self.axle.add_attr(self.carttrans)
            self.axle.set_color(0.1, 1, 1)
            self.viewer.add_geom(self.axle)
  
            # Make another circle on the top of the pole
            self.pole_bob = rendering.make_circle(polewidth/2)
            self.pole_bob_trans = rendering.Transform()
            self.pole_bob.add_attr(self.pole_bob_trans)
            self.pole_bob.add_attr(self.poletrans)
            self.pole_bob.add_attr(self.carttrans)
            self.pole_bob.set_color(0, 0, 0)
            self.viewer.add_geom(self.pole_bob)

            self.wheel_l = rendering.make_circle(cartheight/4)
            self.wheel_r = rendering.make_circle(cartheight/4)
            self.wheeltrans_l = rendering.Transform(translation=(-cartwidth/2, -cartheight/2))
            self.wheeltrans_r = rendering.Transform(translation=(cartwidth/2, -cartheight/2))
            self.wheel_l.add_attr(self.wheeltrans_l)
            self.wheel_l.add_attr(self.carttrans)
            self.wheel_r.add_attr(self.wheeltrans_r)
            self.wheel_r.add_attr(self.carttrans)
            self.wheel_l.set_color(0, 0, 0)  # Black, (B, G, R)
            self.wheel_r.set_color(0, 0, 0)  # Black, (B, G, R)
            self.viewer.add_geom(self.wheel_l)
            self.viewer.add_geom(self.wheel_r)

            self.track = rendering.Line((screen_width/2 - self.x_threshold*scale,carty - cartheight/2 - cartheight/4),
              (screen_width/2 + self.x_threshold*scale,carty - cartheight/2 - cartheight/4))
            self.track.set_color(0,0,0)
            self.viewer.add_geom(self.track)

        if self.state is None: return None

        x = self.state
        cartx = x[0]*scale+screen_width/2.0 # MIDDLE OF CART
        self.carttrans.set_translation(cartx, carty)
        self.poletrans.set_rotation(x[2])
        self.pole_bob_trans.set_translation(-self.l*np.sin(x[2]), self.l*np.cos(x[2]))

        return self.viewer.render(return_rgb_array = mode=='rgb_array')



from matplotlib import pyplot as plt
import networkx as nx
import numpy as np
import sys
sys.path.append('../domain/')
sys.path.append('vis')
from domain.config import games

def viewInd(ind, taskName):
  env = games[taskName]
  if isinstance(ind, str):
    ind = np.loadtxt(ind, delimiter=',') 
    wMat = ind[:,:-1]
    aVec = ind[:,-1]
  else:
    wMat = ind.wMat
    aVec = np.zeros((np.shape(wMat)[0]))  
  print('# of Connections in ANN: ', np.sum(wMat!=0))
    
  # Create Graph
  nIn = env.input_size+1 # bias
  nOut= env.output_size
  G, layer= ind2graph(wMat, nIn, nOut)
  pos = getNodeCoord(G,layer,taskName)
    
  # Draw Graph
  fig = plt.figure(figsize=(10,10), dpi=100)
  ax = fig.add_subplot(111)
  drawEdge(G, pos, wMat, layer)
  nx.draw_networkx_nodes(G,pos,\
    node_color='lightblue',node_shape='o',\
    cmap='terrain',vmin=0,vmax=6)
  drawNodeLabels(G,pos,aVec) 
  labelInOut(pos,env)
  
  plt.tick_params(
    axis='both',       # changes apply to the x-axis
    which='both',      # both major and minor ticks are affected
    bottom=False,      # ticks along the bottom edge are off
    top=False,         # ticks along the top edge are off
    left=False,
    labelleft=False,
    labelbottom=False) # labels along the bottom edge are off
    
  return fig, ax


def ind2graph(wMat, nIn, nOut):
    hMat = wMat[nIn:-nOut,nIn:-nOut]
    hLay = getLayer(hMat)+1

    if len(hLay) > 0:
      lastLayer = max(hLay)+1
    else:
      lastLayer = 1
    L = np.r_[np.zeros(nIn), hLay, np.full((nOut),lastLayer) ]

    layer = L
    order = layer.argsort()
    layer = layer[order]

    wMat = wMat[np.ix_(order,order)]
    nLayer = layer[-1]

    # Convert wMat to Full Network Graph
    rows, cols = np.where(wMat != 0)
    edges = zip(rows.tolist(), cols.tolist())
    G = nx.DiGraph()
    G.add_edges_from(edges)
    return G, layer


def getNodeCoord(G,layer,taskName):
    env = games[taskName]

    # Calculate positions of input and output
    nIn  = env.input_size+1
    nOut = env.output_size 
    nNode= len(G.nodes)
    fixed_pos = np.empty((nNode,2))
    fixed_nodes = np.r_[np.arange(0,nIn),np.arange(nNode-nOut,nNode)]

    # Set Figure dimensions
    fig_wide = 10
    fig_long = 5

    # Assign x and y coordinates per layer
    x = np.ones((1,nNode))*layer # Assign x coord by layer
    x = (x/np.max(x))*fig_wide # Normalize

    _, nPerLayer = np.unique(layer, return_counts=True)

    y = cLinspace(-2,fig_long+2,nPerLayer[0])
    for i in range(1,len(nPerLayer)):
      if i%2 == 0:
        y = np.r_[y,cLinspace(0,fig_long,nPerLayer[i])]
      else:
        y = np.r_[y,cLinspace(-1,fig_long+1,nPerLayer[i])]

    fixed_pos = np.c_[x.T,y.T]
    pos = dict(enumerate(fixed_pos.tolist()))
    
    return pos
  
def labelInOut(pos, env):
  nIn  = env.input_size+1
  nOut = env.output_size 
  nNode= len(pos)
  fixed_nodes = np.r_[np.arange(0,nIn),np.arange(nNode-nOut,nNode)]

  if len(env.in_out_labels)>0:
    stateLabels = ['bias'] + env.in_out_labels
    labelDict = {}
  for i in range(len(stateLabels)):
        labelDict[fixed_nodes[i]] = stateLabels[i]
      
  for i in range(nIn):
    plt.annotate(labelDict[i], xy=(pos[i][0]-0.5, pos[i][1]), xytext=(pos[i][0]-2.5, pos[i][1]-0.5),\
               arrowprops=dict(arrowstyle="->",color='k',connectionstyle="angle"))

  for i in range(nNode-nOut,nNode):
    plt.annotate(labelDict[i], xy=(pos[i][0]+0.1, pos[i][1]), xytext=(pos[i][0]+1.5, pos[i][1]+1.0),\
               arrowprops=dict(arrowstyle="<-",color='k',connectionstyle="angle"))

    
def drawNodeLabels(G, pos, aVec):  
  actLabel = np.array((['','( + )','(0/1)','(sin)','(gau)','(tanh)',\
                        '(sig)','( - )', '(abs)','(relu)','(cos)']))
  listLabel = actLabel[aVec.astype(int)]  
  label = dict(enumerate(listLabel))
  nx.draw_networkx_labels(G,pos,labels=label)  
  
  
def drawEdge(G, pos, wMat, layer):
    wMat[np.isnan(wMat)]=0
    # Organize edges by layer
    _, nPerLayer = np.unique(layer, return_counts=True)
    edgeLayer = []
    layBord = np.cumsum(nPerLayer)
    for i in range(0,len(layBord)):
      tmpMat = np.copy(wMat)
      start = layBord[-i]
      end = layBord[-i+1]
      tmpMat[:,:start] *= 0
      tmpMat[:,end:] *= 0
      rows, cols = np.where(tmpMat != 0)
      edges = zip(rows.tolist(), cols.tolist())
      edgeLayer.append(nx.DiGraph())
      edgeLayer[-1].add_edges_from(edges)
    edgeLayer.append(edgeLayer.pop(0)) # move first layer to correct position

    # Layer Colors
    for i in range(len(edgeLayer)):
      C = [i/len(edgeLayer)] * len(edgeLayer[i].edges) 
      nx.draw_networkx_edges(G,pos,edgelist=edgeLayer[i].edges,\
        alpha=.75,width=1.0,edge_color=C,edge_cmap=plt.cm.viridis,\
        edge_vmin=0.0, edge_vmax=1.0,arrowsize=8)

def getLayer(wMat):
  '''
  Traverse wMat by row, collecting layer of all nodes that connect to you (X).
  Your layer is max(X)+1
  '''
  wMat[np.isnan(wMat)] = 0  
  wMat[wMat!=0]=1
  nNode = np.shape(wMat)[0]
  layer = np.zeros((nNode))
  while(True): # Loop until sorting doesn't help any more
    prevOrder = np.copy(layer)
    for curr in range(nNode):
      srcLayer=np.zeros((nNode))
      for src in range(nNode):
        srcLayer[src] = layer[src]*wMat[src,curr]   
      layer[curr] = np.max(srcLayer)+1    
    if all(prevOrder==layer):
        break
  return layer-1

def cLinspace(start,end,N):
  if N == 1:
    return np.mean([start,end])
  else:
    return np.linspace(start,end,N)

def lload(fileName):
  return np.loadtxt(fileName, delimiter=',') 


from collections import namedtuple
import numpy as np

Game = namedtuple('Game', ['env_name', 'time_factor', 'actionSelect',
  'input_size', 'output_size', 'layers', 'i_act', 'h_act',
  'o_act', 'weightCap','noise_bias','output_noise','max_episode_length','in_out_labels'])

games = {}


# -- Car Racing  --------------------------------------------------------- -- #

# > 32 latent vectors (includes past frames)
vae_racing_stack = Game(env_name='VAERacingStack-v0',
  actionSelect='all', # all, soft, hard
  input_size=32,
  output_size=3,
  time_factor=0,
  layers=[10, 0],
  i_act=np.full(32,1),
  h_act=[1,2,3,4,5,6,7,8,9,10],
  o_act=np.full(3,1),
  weightCap = 2.0,
  noise_bias=0.0,
  max_episode_length = 500,
  output_noise=[False, False, False],
  in_out_labels = ['latent01','latent02','latent03','latent04','latent05',\
                   'latent06','latent07','latent08','latent09','latent10',\
                   'latent11','latent12','latent13','latent14','latent15',\
                   'latent16','latent17','latent18','latent19','latent20',\
                   'latent21','latent22','latent23','latent24','latent25',\
                   'latent26','latent27','latent28','latent29','latent30',\
                   'latent31','latent32','steer'   ,'gas'     ,'brakes']
)
games['vae_racing_stack'] = vae_racing_stack

# > 16 latent vectors (current frame only)
vae_racing = vae_racing_stack._replace(\
  env_name='VAERacing-v0', input_size=16, i_act=np.full(16,1),\
    in_out_labels = ['latent01','latent02','latent03','latent04','latent05',\
                     'latent06','latent07','latent08','latent09','latent10',\
                     'latent11','latent12','latent13','latent14','latent15',\
                     'latent16','steer'   ,'gas'     ,'brakes']  )
games['vae_racing'] = vae_racing


# -- Digit Classification ------------------------------------------------ -- #

# > Scikit learn digits data set
classify = Game(env_name='Classify_digits',
  actionSelect='softmax', # all, soft, hard
  input_size=64,
  output_size=10,
  time_factor=0,
  layers=[128,9],
  i_act=np.full(64,1),
  h_act=[1,3,4,5,6,7,8,9,10], # No step function
  o_act=np.full(10,1),
  weightCap = 2.0,
  noise_bias=0.0,
  output_noise=[False, False, False],
  max_episode_length = 0,
  in_out_labels = []
)
L = [list(range(1, classify.input_size)),\
     list(range(0, classify.output_size))]
label = [item for sublist in L for item in sublist]
classify = classify._replace(in_out_labels=label)
games['digits'] = classify


# > MNIST [28x28] data set
mnist784 = classify._replace(\
  env_name='Classify_mnist784', input_size=784, i_act=np.full(784,1))
L = [list(range(1, mnist784.input_size)),\
     list(range(0, mnist784.output_size))]
label = [item for sublist in L for item in sublist]
mnist784 = mnist784._replace(in_out_labels=label)
games['mnist784'] = mnist784

# > MNIST [16x16] data set
mnist256 = classify._replace(\
  env_name='Classify_mnist256', input_size=256, i_act=np.full(256,1))
L = [list(range(1, mnist256.input_size)),\
     list(range(0, mnist256.output_size))]
label = [item for sublist in L for item in sublist]
mnist256 = mnist256._replace(in_out_labels=label)
games['mnist256'] = mnist256


# -- Cart-pole Swingup --------------------------------------------------- -- #

# > Slower reaction speed
cartpole_swingup = Game(env_name='CartPoleSwingUp_Hard',
  actionSelect='all', # all, soft, hard
  input_size=5,
  output_size=1,
  time_factor=0,
  layers=[5, 5],
  i_act=np.full(5,1),
  h_act=[1,2,3,4,5,6,7,8,9,10],
  o_act=np.full(1,1),
  weightCap = 2.0,
  noise_bias=0.0,
  output_noise=[False, False, False],
  max_episode_length = 200,
  in_out_labels = ['x','x_dot','cos(theta)','sin(theta)','theta_dot',
                   'force']
)
games['swingup_hard'] = cartpole_swingup

# > Normal reaction speed
cartpole_swingup = cartpole_swingup._replace(\
    env_name='CartPoleSwingUp', max_episode_length=1000)
games['swingup'] = cartpole_swingup


# -- Bipedal Walker ------------------------------------------------------ -- #

# > Flat terrain
biped = Game(env_name='BipedalWalker-v2',
  actionSelect='all', # all, soft, hard
  input_size=24,
  output_size=4,
  time_factor=0,
  layers=[40, 40],
  i_act=np.full(24,1),
  h_act=[1,2,3,4,5,6,7,8,9,10],
  o_act=np.full(4,1),
  weightCap = 2.0,
  noise_bias=0.0,
  output_noise=[False, False, False],
  max_episode_length = 400,
  in_out_labels = [
  'hull_angle','hull_vel_angle','vel_x','vel_y',
  'hip1_angle','hip1_speed','knee1_angle','knee1_speed','leg1_contact',
  'hip2_angle','hip2_speed','knee2_angle','knee2_speed','leg2_contact',
  'lidar_0','lidar_1','lidar_2','lidar_3','lidar_4',
  'lidar_5','lidar_6','lidar_7','lidar_8','lidar_9',
  'hip_1','knee_1','hip_2','knee_2']
)
games['biped'] = biped

# > Hilly Terrain
bipedmed = biped._replace(env_name='BipedalWalkerMedium-v2')
games['bipedmedium'] = bipedmed

# > Obstacles, hills, and pits
bipedhard = biped._replace(env_name='BipedalWalkerHardcore-v2')
games['bipedhard'] = bipedhard


# -- Bullet -------------------------------------------------------------- -- #

# > Quadruped ant
bullet_ant = Game(env_name='AntBulletEnv-v0',
  actionSelect='all', # all, soft, hard
  input_size=28,
  output_size=8,
  layers=[64, 32],
  time_factor=1000,
  i_act=np.full(28,1),
  h_act=[1,2,3,4,5,6,7,8,9,10],
  o_act=np.full(8,1),
  weightCap = 2.0,
  noise_bias=0.0,
  output_noise=[False, False, True],
  max_episode_length = 1000,
  in_out_labels = []
)
games['bullet_ant'] = bullet_ant




import os
import numpy as np
import copy
from .ann import exportNet

class DataGatherer():
  """Data recorder for NEAT algorithm
  """
  def __init__(self, filename, hyp): 
    """
    Args:
      filename - (string) - path+prefix of file output destination
      hyp      - (dict)   - algorithm hyperparameters
    """
    self.filename = filename # File name path + prefix
    self.p = hyp
    
    # Initialize empty fields
    self.elite = []
    self.best = []
    self.bestFitVec = []
    self.spec_fit = []
    self.field = ['x_scale','fit_med','fit_max','fit_top',\
                  'node_med','conn_med',\
                  'elite','best']

    if self.p['alg_probMoo'] > 0:
      self.objVals = np.array([])

    for f in self.field[:-2]:
      exec('self.' + f + ' = np.array([])')
      #e.g. self.fit_max   = np.array([]) 

    self.newBest = False

  def gatherData(self, pop, species):
    """Collect and stores run data
    """

    # Readability
    fitness = [ind.fitness for ind in pop]
    nodes = np.asarray([np.shape(ind.node)[1] for ind in pop])
    conns = np.asarray([ind.nConn for ind in pop])
    
    # --- Evaluation Scale ---------------------------------------------------
    if len(self.x_scale) == 0:
      self.x_scale = np.append(self.x_scale, len(pop))
    else:
      self.x_scale = np.append(self.x_scale, self.x_scale[-1]+len(pop))
    # ------------------------------------------------------------------------ 

    
    # --- Best Individual ----------------------------------------------------
    self.elite.append(pop[np.argmax(fitness)])
    if len(self.best) == 0:
      self.best = copy.deepcopy(self.elite)
    elif (self.elite[-1].fitness > self.best[-1].fitness):
      self.best = np.append(self.best,copy.deepcopy(self.elite[-1]))
      self.newBest = True
    else:
      self.best = np.append(self.best,copy.deepcopy(self.best[-1]))   
      self.newBest = False
    # ------------------------------------------------------------------------ 

    
    # --- Generation fit/complexity stats ------------------------------------ 
    self.node_med = np.append(self.node_med,np.median(nodes))
    self.conn_med = np.append(self.conn_med,np.median(conns))
    self.fit_med  = np.append(self.fit_med, np.median(fitness))
    self.fit_max  = np.append(self.fit_max,  self.elite[-1].fitness)
    self.fit_top  = np.append(self.fit_top,  self.best[-1].fitness)
    # ------------------------------------------------------------------------ 


    # --- MOO Fronts ---------------------------------------------------------
    if self.p['alg_probMoo'] > 0:
      if len(self.objVals) == 0:
        self.objVals = np.c_[fitness,conns]
      else:
        self.objVals = np.c_[self.objVals, np.c_[fitness,conns]]
    # ------------------------------------------------------------------------ 

    
    # --- Species Stats ------------------------------------------------------
    if self.p['alg_speciate'] == 'neat':
      specFit = np.empty((2,0))
      #print('# of Species: ', len(species))
      for iSpec in range(len(species)):
        for ind in species[iSpec].members:
          tmp = np.array((iSpec,ind.fitness))
          specFit = np.c_[specFit,tmp]
      self.spec_fit = specFit
    # ------------------------------------------------------------------------ 


  def display(self):
    """Console output for each generation
    """
    return    "|---| Elite Fit: " + '{:.2f}'.format(self.fit_max[-1]) \
         + " \t|---| Best Fit:  "  + '{:.2f}'.format(self.fit_top[-1])

  def save(self, gen=(-1)):
    """Save algorithm stats to disk
    """
    ''' Save data to disk '''
    filename = self.filename
    pref = 'log/' + filename

    # --- Generation fit/complexity stats ------------------------------------ 
    gStatLabel = ['x_scale',\
                  'fit_med','fit_max','fit_top','node_med','conn_med']
    genStats = np.empty((len(self.x_scale),0))
    for i in range(len(gStatLabel)):
      #e.g.         self.    fit_max          [:,None]
      evalString = 'self.' + gStatLabel[i] + '[:,None]'
      genStats = np.hstack((genStats, eval(evalString)))
    lsave(pref + '_stats.out', genStats)
    # ------------------------------------------------------------------------ 


    # --- Best Individual ----------------------------------------------------
    wMat = self.best[gen].wMat
    aVec = self.best[gen].aVec
    exportNet(pref + '_best.out',wMat,aVec)
    
    if gen > 1:
      folder = 'log/' + filename + '_best/'
      if not os.path.exists(folder):
        os.makedirs(folder)
      exportNet(folder + str(gen).zfill(4) +'.out',wMat,aVec)
    # ------------------------------------------------------------------------


    # --- Species Stats ------------------------------------------------------
    if self.p['alg_speciate'] == 'neat':
      lsave(pref + '_spec.out', self.spec_fit)
    # ------------------------------------------------------------------------


    # --- MOO Fronts ---------------------------------------------------------
    if self.p['alg_probMoo'] > 0:
      lsave(pref + '_objVals.out',self.objVals)
    # ------------------------------------------------------------------------

  def savePop(self,pop,filename):
    """Save all individuals in population as numpy arrays
    """
    folder = 'log/' + filename + '_pop/'
    if not os.path.exists(folder):
      os.makedirs(folder)

    for i in range(len(pop)):
      exportNet(folder+'ind_'+str(i)+'.out', pop[i].wMat, pop[i].aVec)

def lsave(filename, data):
  """Short hand for numpy save with csv and float precision defaults
  """
  np.savetxt(filename, data, delimiter=',',fmt='%1.2e')








import numpy as np
from .utils import *

class Species():
  """Species class, only contains fields: all methods belong to the NEAT class.
  Note: All 'species' related functions are part of the Neat class, though defined in this file.
  """

  def __init__(self,seed):
    """Intialize species around a seed
    Args:
      seed - (Ind) - individual which anchors seed in compatibility space

    Attributes:
      seed       - (Ind)   - individual who acts center of species
      members    - [Ind]   - individuals in species
      bestInd    - (Ind)   - highest fitness individual ever found in species
      bestFit    - (float) - highest fitness ever found in species
      lastImp    - (int)   - generations since a new best individual was found
      nOffspring - (int)   - new individuals to create this generation
    """
    self.seed = seed      # Seed is type Ind
    self.members = [seed] # All inds in species
    self.bestInd = seed
    self.bestFit = seed.fitness
    self.lastImp = 0
    self.nOffspring = []

def speciate(self):  
  """Divides population into species and assigns each a number of offspring/
  """
  # Readbility
  p = self.p
  pop = self.pop
  species = self.species

  if p['alg_speciate'] == 'neat':
    # Adjust species threshold to track desired number of species
    if len(species) > p['spec_target']:
      p['spec_thresh'] += p['spec_compatMod']

    if len(species) < p['spec_target']:
      p['spec_thresh'] -= p['spec_compatMod']

    if p['spec_thresh'] < p['spec_threshMin']:
      p['spec_thresh'] = p['spec_threshMin']

    species, pop = self.assignSpecies  (species, pop, p)
    species      = self.assignOffspring(species, pop, p)

  elif p['alg_speciate'] == "none": 
    # Recombination takes a species, when there is no species we dump the whole population into one species that is awarded all offspring
    species = [Species(pop[0])]
    species[0].nOffspring = p['popSize']
    for ind in pop:
      ind.species = 0
    species[0].members = pop

  # Update
  self.p = p
  self.pop = pop
  self.species = species

def assignSpecies(self, species, pop, p):
  """Assigns each member of the population to a species.
  Fills each species class with nearests members, assigns a species Id to each
  individual for record keeping

  Args:
    species - (Species) - Previous generation's species
      .seed       - (Ind) - center of species
    pop     - [Ind]     - unassigned individuals
    p       - (Dict)    - algorithm hyperparameters

  Returns:
    species - (Species) - This generation's species
      .seed       - (Ind) - center of species
      .members    - [Ind] - parent population
    pop     - [Ind]     - individuals with species ID assigned

  """

  # Get Previous Seeds
  if len(self.species) == 0:
    # Create new species if none exist
    species = [Species(pop[0])]
    species[0].nOffspring = p['popSize']
    iSpec = 0
  else:
    # Remove existing members
    for iSpec in range(len(species)):
      species[iSpec].members = []

  # Assign members of population to first species within compat distance
  for i in range(len(pop)):
    assigned = False
    for iSpec in range(len(species)):
      ref = np.copy(species[iSpec].seed.conn)
      ind = np.copy(pop[i].conn)
      cDist = self.compatDist(ref,ind)
      if cDist < p['spec_thresh']:
        pop[i].species = iSpec
        species[iSpec].members.append(pop[i])
        assigned = True
        break

    # If no seed is close enough, start your own species
    if not assigned:
      pop[i].species = iSpec+1
      species.append(Species(pop[i]))

  return species, pop

def assignOffspring(self, species, pop, p):
  """Assigns number of offspring to each species based on fitness sharing.
  NOTE: Ordinal rather than the cardinal fitness of canonical NEAT is used.

  Args:
    species - (Species) - this generation's species
      .members    - [Ind]   - individuals in species
    pop     - [Ind]     - individuals with species assigned
      .fitness    - (float) - performance on task (higher is better)
    p       - (Dict)    - algorithm hyperparameters

  Returns:
    species - (Species) - This generation's species
      .nOffspring - (int) - number of children to produce
  """

  nSpecies = len(species)
  if nSpecies == 1:
    species[0].offspring = p['popSize']
  else:
    # -- Fitness Sharing
    # Rank all individuals
    popFit = np.asarray([ind.fitness for ind in pop])
    popRank = tiedRank(popFit)
    if p['select_rankWeight'] == 'exp':
      rankScore = 1/popRank
    elif p['select_rankWeight'] == 'lin':
      rankScore = 1+abs(popRank-len(popRank))
    else:
      print("Invalid rank weighting (using linear)")
      rankScore = 1+abs(popRank-len(popRank))
    specId = np.asarray([ind.species for ind in pop])

    # Best and Average Fitness of Each Species
    speciesFit = np.zeros((nSpecies,1))
    speciesTop = np.zeros((nSpecies,1))
    for iSpec in range(nSpecies):
      if not np.any(specId==iSpec):
        speciesFit[iSpec] = 0
      else:
        speciesFit[iSpec] = np.mean(rankScore[specId==iSpec])
        speciesTop[iSpec] = np.max(popFit[specId==iSpec])

        # Did the species improve?
        if speciesTop[iSpec] > species[iSpec].bestFit:
          species[iSpec].bestFit = speciesTop[iSpec]
          bestId = np.argmax(popFit[specId==iSpec])
          species[iSpec].bestInd = species[iSpec].members[bestId]
          species[iSpec].lastImp = 0
        else:
          species[iSpec].lastImp += 1

        # Stagnant species don't recieve species fitness
        if species[iSpec].lastImp > p['spec_dropOffAge']:
          speciesFit[iSpec] = 0
          
    # -- Assign Offspring
    if sum(speciesFit) == 0:
      speciesFit = np.ones((nSpecies,1))
      print("WARN: Entire population stagnant, continuing without extinction")
      
    offspring = bestIntSplit(speciesFit, p['popSize'])
    for iSpec in range(nSpecies):
      species[iSpec].nOffspring = offspring[iSpec]
      
  # Extinction    
  species[:] = [s for s in species if s.nOffspring != 0]

  return species

def compatDist(self, ref, ind):
  """Calculate 'compatiblity distance' between to genomes

  Args:
    ref - (np_array) -  reference genome connection genes
          [5 X nUniqueGenes]
          [0,:] == Innovation Number (unique Id)
          [3,:] == Weight Value
    ind - (np_array) -  genome being compared
          [5 X nUniqueGenes]
          [0,:] == Innovation Number (unique Id)
          [3,:] == Weight Value

  Returns:
    dist - (float) - compatibility distance between genomes
  """

  # Find matching genes
  IA, IB = quickINTersect(ind[0,:].astype(int),ref[0,:].astype(int))          
  
  # Calculate raw genome distances
  ind[3,np.isnan(ind[3,:])] = 0
  ref[3,np.isnan(ref[3,:])] = 0
  weightDiff = abs(ind[3,IA] - ref[3,IB])
  geneDiff   = sum(np.invert(IA)) + sum(np.invert(IB))

  # Normalize and take weighted sum
  nInitial = self.p['ann_nInput'] + self.p['ann_nOutput']
  longestGenome = max(len(IA),len(IB)) - nInitial
  weightDiff = np.mean(weightDiff)
  geneDiff   = geneDiff   / (1+longestGenome)

  dist = geneDiff   * self.p['spec_geneCoef']      \
       + weightDiff * self.p['spec_weightCoef']  
  return dist



import numpy as np


# -- ANN Ordering -------------------------------------------------------- -- #

def getNodeOrder(nodeG,connG):
  """Builds connection matrix from genome through topological sorting.

  Args:
    nodeG - (np_array) - node genes
            [3 X nUniqueGenes]
            [0,:] == Node Id
            [1,:] == Type (1=input, 2=output 3=hidden 4=bias)
            [2,:] == Activation function (as int)

    connG - (np_array) - connection genes
            [5 X nUniqueGenes] 
            [0,:] == Innovation Number (unique Id)
            [1,:] == Source Node Id
            [2,:] == Destination Node Id
            [3,:] == Weight Value
            [4,:] == Enabled?  

  Returns:
    Q    - [int]      - sorted node order as indices
    wMat - (np_array) - ordered weight matrix
           [N X N]

    OR

    False, False      - if cycle is found

  Todo:
    * setdiff1d is slow, as all numbers are positive ints is there a
      better way to do with indexing tricks (as in quickINTersect)?
  """
  conn = np.copy(connG)
  node = np.copy(nodeG)
  nIns = len(node[0,node[1,:] == 1]) + len(node[0,node[1,:] == 4])
  nOuts = len(node[0,node[1,:] == 2])
  
  # Create connection and initial weight matrices
  conn[3,conn[4,:]==0] = np.nan # disabled but still connected
  src  = conn[1,:].astype(int)
  dest = conn[2,:].astype(int)
  
  lookup = node[0,:].astype(int)
  for i in range(len(lookup)): # Can we vectorize this?
    src[np.where(src==lookup[i])] = i
    dest[np.where(dest==lookup[i])] = i
  
  wMat = np.zeros((np.shape(node)[1],np.shape(node)[1]))
  wMat[src,dest] = conn[3,:]
  connMat = wMat[nIns+nOuts:,nIns+nOuts:]
  connMat[connMat!=0] = 1

  # Topological Sort of Hidden Nodes
  edge_in = np.sum(connMat,axis=0)
  Q = np.where(edge_in==0)[0]  # Start with nodes with no incoming connections
  for i in range(len(connMat)):
    if (len(Q) == 0) or (i >= len(Q)):
      Q = []
      return False, False # Cycle found, can't sort
    edge_out = connMat[Q[i],:]
    edge_in  = edge_in - edge_out # Remove nodes' conns from total
    nextNodes = np.setdiff1d(np.where(edge_in==0)[0], Q)
    Q = np.hstack((Q,nextNodes))

    if sum(edge_in) == 0:
      break
  
  # Add In and outs back and reorder wMat according to sort
  Q += nIns+nOuts
  Q = np.r_[lookup[:nIns], Q, lookup[nIns:nIns+nOuts]]
  wMat = wMat[np.ix_(Q,Q)]
  
  return Q, wMat

def getLayer(wMat):
  """Get layer of each node in weight matrix
  Traverse wMat by row, collecting layer of all nodes that connect to you (X).
  Your layer is max(X)+1. Input and output nodes are ignored and assigned layer
  0 and max(X)+1 at the end.

  Args:
    wMat  - (np_array) - ordered weight matrix
           [N X N]

  Returns:
    layer - [int]      - layer # of each node

  Todo:
    * With very large networks this might be a performance sink -- especially, 
    given that this happen in the serial part of the algorithm. There is
    probably a more clever way to do this given the adjacency matrix.
  """
  wMat[np.isnan(wMat)] = 0  
  wMat[wMat!=0]=1
  nNode = np.shape(wMat)[0]
  layer = np.zeros((nNode))
  while (True): # Loop until sorting is stable
    prevOrder = np.copy(layer)
    for curr in range(nNode):
      srcLayer=np.zeros((nNode))
      for src in range(nNode):
        srcLayer[src] = layer[src]*wMat[src,curr]   
      layer[curr] = np.max(srcLayer)+1    
    if all(prevOrder==layer):
      break
  return layer-1


# -- ANN Activation ------------------------------------------------------ -- #

def act(weights, aVec, nInput, nOutput, inPattern):
  """Returns FFANN output given a single input pattern
  If the variable weights is a vector it is turned into a square weight matrix.
  
  Allows the network to return the result of several samples at once if given a matrix instead of a vector of inputs:
      Dim 0 : individual samples
      Dim 1 : dimensionality of pattern (# of inputs)

  Args:
    weights   - (np_array) - ordered weight matrix or vector
                [N X N] or [N**2]
    aVec      - (np_array) - activation function of each node 
                [N X 1]    - stored as ints (see applyAct in ann.py)
    nInput    - (int)      - number of input nodes
    nOutput   - (int)      - number of output nodes
    inPattern - (np_array) - input activation
                [1 X nInput] or [nSamples X nInput]

  Returns:
    output    - (np_array) - output activation
                [1 X nOutput] or [nSamples X nOutput]
  """
  # Turn weight vector into weight matrix
  if np.ndim(weights) < 2:
      nNodes = int(np.sqrt(np.shape(weights)[0]))
      wMat = np.reshape(weights, (nNodes, nNodes))
  else:
      nNodes = np.shape(weights)[0]
      wMat = weights
  wMat[np.isnan(wMat)]=0

  # Vectorize input
  if np.ndim(inPattern) > 1:
      nSamples = np.shape(inPattern)[0]
  else:
      nSamples = 1

  # Run input pattern through ANN    
  nodeAct  = np.zeros((nSamples,nNodes))
  nodeAct[:,0] = 1 # Bias activation
  nodeAct[:,1:nInput+1] = inPattern

  # Propagate signal through hidden to output nodes
  iNode = nInput+1
  for iNode in range(nInput+1,nNodes):
      rawAct = np.dot(nodeAct, wMat[:,iNode]).squeeze()
      nodeAct[:,iNode] = applyAct(aVec[iNode], rawAct) 
      #print(nodeAct)
  output = nodeAct[:,-nOutput:]   
  return output

def applyAct(actId, x):
  """Returns value after an activation function is applied
  Lookup table to allow activations to be stored in numpy arrays

  case 1  -- Linear
  case 2  -- Unsigned Step Function
  case 3  -- Sin
  case 4  -- Gausian with mean 0 and sigma 1
  case 5  -- Hyperbolic Tangent [tanh] (signed)
  case 6  -- Sigmoid unsigned [1 / (1 + exp(-x))]
  case 7  -- Inverse
  case 8  -- Absolute Value
  case 9  -- Relu
  case 10 -- Cosine
  case 11 -- Squared

  Args:
    actId   - (int)   - key to look up table
    x       - (???)   - value to be input into activation
              [? X ?] - any type or dimensionality

  Returns:
    output  - (float) - value after activation is applied
              [? X ?] - same dimensionality as input
  """
  if actId == 1:   # Linear
    value = x

  if actId == 2:   # Unsigned Step Function
    value = 1.0*(x>0.0)
    #value = (np.tanh(50*x/2.0) + 1.0)/2.0

  elif actId == 3: # Sin
    value = np.sin(np.pi*x) 

  elif actId == 4: # Gaussian with mean 0 and sigma 1
    value = np.exp(-np.multiply(x, x) / 2.0)

  elif actId == 5: # Hyperbolic Tangent (signed)
    value = np.tanh(x)     

  elif actId == 6: # Sigmoid (unsigned)
    value = (np.tanh(x/2.0) + 1.0)/2.0

  elif actId == 7: # Inverse
    value = -x

  elif actId == 8: # Absolute Value
    value = abs(x)   
    
  elif actId == 9: # Relu
    value = np.maximum(0, x)   

  elif actId == 10: # Cosine
    value = np.cos(np.pi*x)

  elif actId == 11: # Squared
    value = x**2
    
  else:
    value = x

  return value


# -- Action Selection ---------------------------------------------------- -- #

def selectAct(action, actSelect):  
  """Selects action based on vector of actions

    Single Action:
    - Hard: a single action is chosen based on the highest index
    - Prob: a single action is chosen probablistically with higher values
            more likely to be chosen

    We aren't selecting a single action:
    - Softmax: a softmax normalized distribution of values is returned
    - Default: all actions are returned 

  Args:
    action   - (np_array) - vector weighting each possible action
                [N X 1]

  Returns:
    i         - (int) or (np_array)     - chosen index
                         [N X 1]
  """  
  if actSelect == 'softmax':
    action = softmax(action)
  elif actSelect == 'prob':
    action = weightedRandom(np.sum(action,axis=0))
  else:
    action = action.flatten()
  return action

def softmax(x):
    """Compute softmax values for each sets of scores in x.
    Assumes: [samples x dims]

    Args:
      x - (np_array) - unnormalized values
          [samples x dims]

    Returns:
      softmax - (np_array) - softmax normalized in dim 1
    
    Todo: Untangle all the transposes...    
    """    
    if x.ndim == 1:
      e_x = np.exp(x - np.max(x))
      return e_x / e_x.sum(axis=0)
    else:
      e_x = np.exp(x.T - np.max(x,axis=1))
      return (e_x / e_x.sum(axis=0)).T

def weightedRandom(weights):
  """Returns random index, with each choices chance weighted
  Args:
    weights   - (np_array) - weighting of each choice
                [N X 1]

  Returns:
    i         - (int)      - chosen index
  """
  minVal = np.min(weights)
  weights = weights - minVal # handle negative vals
  cumVal = np.cumsum(weights)
  pick = np.random.uniform(0, cumVal[-1])
  for i in range(len(weights)):
    if cumVal[i] >= pick:
      return i
        

# -- File I/O ------------------------------------------------------------ -- #

def exportNet(filename,wMat, aVec):
  indMat = np.c_[wMat,aVec]
  np.savetxt(filename, indMat, delimiter=',',fmt='%1.2e')

def importNet(fileName):
  ind = np.loadtxt(fileName, delimiter=',')
  wMat = ind[:,:-1]     # Weight Matrix
  aVec = ind[:,-1]      # Activation functions

  # Create weight key
  wVec = wMat.flatten()
  wVec[np.isnan(wVec)]=0
  wKey = np.where(wVec!=0)[0] 

  return wVec, aVec, wKey




import numpy as np
import itertools
from .ind import Ind
from .utils import *


def evolvePop(self):
  """ Evolves new population from existing species.
  Wrapper which calls 'recombine' on every species and combines all offspring into a new population. When speciation is not used, the entire population is treated as a single species.
  """
  newPop = []
  for i in range(len(self.species)):
    children, self.innov = self.recombine(self.species[i],\
                           self.innov, self.gen)
    newPop.append(children)
  self.pop = list(itertools.chain.from_iterable(newPop))   

def recombine(self, species, innov, gen):
  """ Creates next generation of child solutions from a species

  Procedure:
    ) Sort all individuals by rank
    ) Eliminate lower percentage of individuals from breeding pool
    ) Pass upper percentage of individuals to child population unchanged
    ) Select parents by tournament selection
    ) Produce new population through crossover and mutation

  Args:
      species - (Species) -
        .members    - [Ind] - parent population
        .nOffspring - (int) - number of children to produce
      innov   - (np_array)  - innovation record
                [5 X nUniqueGenes]
                [0,:] == Innovation Number
                [1,:] == Source
                [2,:] == Destination
                [3,:] == New Node?
                [4,:] == Generation evolved
      gen     - (int) - current generation

  Returns:
      children - [Ind]      - newly created population
      innov   - (np_array)  - updated innovation record

  """
  p = self.p
  nOffspring = int(species.nOffspring)
  pop = species.members
  children = []
 
  # Sort by rank
  pop.sort(key=lambda x: x.rank)

  # Cull  - eliminate worst individuals from breeding pool
  numberToCull = int(np.floor(p['select_cullRatio'] * len(pop)))
  if numberToCull > 0:
    pop[-numberToCull:] = []     

  # Elitism - keep best individuals unchanged
  nElites = int(np.floor(len(pop)*p['select_eliteRatio']))
  for i in range(nElites):
    children.append(pop[i])
    nOffspring -= 1

  # Get parent pairs via tournament selection
  # -- As individuals are sorted by fitness, index comparison is 
  # enough. In the case of ties the first individual wins
  parentA = np.random.randint(len(pop),size=(nOffspring,p['select_tournSize']))
  parentB = np.random.randint(len(pop),size=(nOffspring,p['select_tournSize']))
  parents = np.vstack( (np.min(parentA,1), np.min(parentB,1) ) )
  parents = np.sort(parents,axis=0) # Higher fitness parent first    
  
  # Breed child population
  for i in range(nOffspring):  
    if np.random.rand() > p['prob_crossover']:
      # Mutation only: take only highest fit parent
      child, innov = pop[parents[0,i]].createChild(p,innov,gen)
    else:
      # Crossover
      child, innov = pop[parents[0,i]].createChild(p,innov,gen,\
                mate=pop[parents[1,i]])

    child.express()
    children.append(child)      

  return children, innov



from .task_gym import GymTask
from .config import games
import json
import numpy as np

def loadHyp(pFileName, printHyp=False):
  """Loads hyperparameters from .json file
  Args:
      pFileName - (string) - file name of hyperparameter file
      printHyp  - (bool)   - print contents of hyperparameter file to terminal?

  Note: see config/hypkey.txt for detailed hyperparameter description
  """
  with open(pFileName) as data_file: hyp = json.load(data_file)

  # Task hyper parameters
  task = GymTask(games[hyp['task']],paramOnly=True)
  hyp['ann_nInput']   = task.nInput
  hyp['ann_nOutput']  = task.nOutput
  hyp['ann_initAct']  = task.activations[0]
  hyp['ann_absWCap']  = task.absWCap
  hyp['ann_mutSigma'] = task.absWCap * 0.2
  hyp['ann_layers']   = task.layers # if fixed toplogy is used

  if hyp['alg_act'] == 0:
    hyp['ann_actRange'] = task.actRange
  else:
    hyp['ann_actRange'] = np.full_like(task.actRange,hyp['alg_act'])



  if printHyp is True:
    print(json.dumps(hyp, indent=4, sort_keys=True))
  return hyp

def updateHyp(hyp,pFileName=None):
  """Overwrites default hyperparameters with those from second .json file
  """
  if pFileName != None:
    print('\t*** Running with hyperparameters: ', pFileName, '\t***')
    with open(pFileName) as data_file: update = json.load(data_file)
    hyp.update(update)

    # Task hyper parameters
    task = GymTask(games[hyp['task']],paramOnly=True)
    hyp['ann_nInput']   = task.nInput
    hyp['ann_nOutput']  = task.nOutput
    hyp['ann_initAct']  = task.activations[0]
    hyp['ann_absWCap']  = task.absWCap
    hyp['ann_mutSigma'] = task.absWCap * 0.1
    hyp['ann_layers']   = task.layers # if fixed toplogy is used


    if hyp['alg_act'] == 0:
      hyp['ann_actRange'] = task.actRange
    else:
      hyp['ann_actRange'] = np.full_like(task.actRange,hyp['alg_act'])



import glob
from matplotlib import pyplot as plt
import networkx as nx
import numpy as np
import pickle
import warnings
from matplotlib import cm

# Custom libraties
from .lplot import *


def loadPop(popFile):
  file = open(popFile,'rb')
  return pickle.load(file) 


# -- Replicates ---- ----------------------------------------------#
def viewFitFile(fname,val='Fit',label=None,axis=False):
  fig, ax = getAxis(axis)
  stats = lload(fname)
  x = stats[:,0]
  if val == 'Fit':
    fit = stats[:,[1,2,4]]
    #lplot(x,fit,label=['Min','Med', 'Top'],axis=ax)
    lplot(fit,label=['Min','Med', 'Top'],axis=ax)
  else:
    fit = stats[:,5]
    lplot(x,fit,label=['Conns'],axis=ax)
  return fig, ax
    

  
def viewReps(prefix,label=[],val='Fit', title='Fitness',\
             axis=False, getBest=False):
  fig, ax = getAxis(axis)
  fig.dpi=100
  bestRun = []    
  for pref in prefix:
    statFile = sorted(glob.glob(pref + '*stats.out'))
    if len(statFile) == 0:
      print('ERROR: No files with that prefix found (it is a list?)')
      return False
    
    for i in range(len(statFile)):
      tmp = lload(statFile[i]) 
      if i == 0:
        x = tmp[:,0]
        if val == 'Conn':
          fitVal = tmp[:,5]
        else: # Fitness
          fitVal = tmp[:,3]
          bestVal = fitVal[-1]
          bestRun.append(statFile[i])
      else:
        if np.shape(tmp)[0] != np.shape(fitVal)[0]:
          print("Incomplete file found, ignoring ", statFile[i], ' and later.')
          break
        
        if val == 'Conn':
          fitVal = np.c_[fitVal,tmp[:,5]]    
        else: # Fitness
          fitVal = np.c_[fitVal,tmp[:,3]]
          if fitVal[-1,-1] > bestVal:
            bestVal = fitVal[-1,-1]
            bestRun[-1] = statFile[i]
    
    x = np.arange(len(x))
    lquart(x,fitVal,axis=ax) # Display Quartiles

  # Legend
  if len(label) > 0:
    newLeg = []
    for i in range(len(label)):
      newLeg.append(label[i])
      newLeg.append('_nolegend_')
      newLeg.append('_nolegend_')
    warnings.filterwarnings("ignore", category=UserWarning)
    plt.gca().legend((newLeg))
  plt.title(title)
  plt.xlabel('Evaluations')
  plt.xlabel('Generations')
  if val == 'Conn':
    plt.ylabel('Median Connections')
  else: # Fitness
    plt.ylabel('Best Fitness Found')

  if getBest is True:
    return fig,ax,bestRun
  else:
    return fig,ax
# -- ------------ -- ----------------------------------------------#

def getAxis(axis):
  if axis is not False:
    ax = axis
    fig = ax.figure.canvas 
  else:
    fig, ax = plt.subplots()
    
  return fig,ax


import logging
import math
import gym
from gym import spaces
from gym.utils import seeding
import numpy as np
import sys
import cv2
import math

class ClassifyEnv(gym.Env):
  """Classification as an unsupervised OpenAI Gym RL problem.
  Includes scikit-learn digits dataset, MNIST dataset
  """

  def __init__(self, trainSet, target):
    """
    Data set is a tuple of 
    [0] input data: [nSamples x nInputs]
    [1] labels:     [nSamples x 1]

    Example data sets are given at the end of this file
    """

    self.t = 0          # Current batch number
    self.t_limit = 0    # Number of batches if you need them
    self.batch   = 1000 # Number of images per batch
    self.seed()
    self.viewer = None

    self.trainSet = trainSet
    self.target   = target

    nInputs = np.shape(trainSet)[1]
    high = np.array([1.0]*nInputs)
    self.action_space = spaces.Box(np.array(0,dtype=np.float32), \
                                   np.array(1,dtype=np.float32))
    self.observation_space = spaces.Box(np.array(0,dtype=np.float32), \
                                   np.array(1,dtype=np.float32))

    self.state = None
    self.trainOrder = None
    self.currIndx = None

  def seed(self, seed=None):
    ''' Randomly select from training set'''
    self.np_random, seed = seeding.np_random(seed)
    return [seed]
  
  def reset(self):
    ''' Initialize State'''    
    #print('Lucky number', np.random.randint(10)) # same randomness?
    self.trainOrder = np.random.permutation(len(self.target))
    self.t = 0 # timestep
    self.currIndx = self.trainOrder[self.t:self.t+self.batch]
    self.state = self.trainSet[self.currIndx,:]
    return self.state
  
  def step(self, action):
    ''' 
    Judge Classification, increment to next batch
    action - [batch x output] - softmax output
    '''
    y = self.target[self.currIndx]
    m = y.shape[0]

    log_likelihood = -np.log(action[range(m),y])
    loss = np.sum(log_likelihood) / m
    reward = -loss

    if self.t_limit > 0: # We are doing batches
      reward *= (1/self.t_limit) # average
      self.t += 1
      done = False
      if self.t >= self.t_limit:
        done = True
      self.currIndx = self.trainOrder[(self.t*self.batch):\
                                      (self.t*self.batch + self.batch)]

      self.state = self.trainSet[self.currIndx,:]
    else:
      done = True

    obs = self.state
    return obs, reward, done, {}


# -- Data Sets ----------------------------------------------------------- -- #

def digit_raw():
  ''' 
  Converts 8x8 scikit digits to 
  [samples x pixels]  ([N X 64])
  '''  
  from sklearn import datasets
  digits = datasets.load_digits()
  z = (digits.images/16)
  z = z.reshape(-1, (64))
  return z, digits.target

def mnist_784():
  ''' 
  Converts 28x28 mnist digits to 
  [samples x pixels]  ([N X 784])
  '''  
  import mnist
  z = (mnist.train_images()/255)
  z = preprocess(z,(28,28))
  z = z.reshape(-1, (784))
  return z, mnist.train_labels()

def mnist_256():
  ''' 
  Converts 28x28 mnist digits to [16x16] 
  [samples x pixels]  ([N X 256])
  '''  
  import mnist
  z = (mnist.train_images()/255)
  z = preprocess(z,(16,16))

  z = z.reshape(-1, (256))
  return z, mnist.train_labels()


def preprocess(img,size, patchCorner=(0,0), patchDim=None, unskew=True):
  """
  Resizes, crops, and unskewes images

  """
  if patchDim == None: patchDim = size
  nImg = np.shape(img)[0]
  procImg  = np.empty((nImg,size[0],size[1]))

  # Unskew and Resize
  if unskew == True:    
    for i in range(nImg):
      procImg[i,:,:] = deskew(cv2.resize(img[i,:,:],size),size)

  # Crop
  cropImg  = np.empty((nImg,patchDim[0],patchDim[1]))
  for i in range(nImg):
    cropImg[i,:,:] = procImg[i,patchCorner[0]:patchCorner[0]+patchDim[0],\
                               patchCorner[1]:patchCorner[1]+patchDim[1]]
  procImg = cropImg

  return procImg

def deskew(image, image_shape, negated=True):
  """
  This method deskwes an image using moments
  :param image: a numpy nd array input image
  :param image_shape: a tuple denoting the image`s shape
  :param negated: a boolean flag telling whether the input image is negated

  :returns: a numpy nd array deskewd image

  source: https://github.com/vsvinayak/mnist-helper
  """
  
  # negate the image
  if not negated:
      image = 255-image
  # calculate the moments of the image
  m = cv2.moments(image)
  if abs(m['mu02']) < 1e-2:
      return image.copy()
  # caclulating the skew
  skew = m['mu11']/m['mu02']
  M = np.float32([[1, skew, -0.5*image_shape[0]*skew], [0,1,0]])
  img = cv2.warpAffine(image, M, image_shape, \
    flags=cv2.WARP_INVERSE_MAP|cv2.INTER_LINEAR)  
  return img



 



import numpy as np
import warnings

def nsga_sort(objVals, returnFronts=False):
  """Returns ranking of objective values based on non-dominated sorting.
  Optionally returns fronts (useful for visualization).
  
  NOTE: Assumes maximization of objective function
   
  Args: 
    objVals - (np_array) - Objective values of each individual
              [nInds X nObjectives]
    
  Returns: 
    rank    - (np_array) - Rank in population of each individual
            int([nIndividuals X 1])
    front   - (np_array) - Pareto front of each individual
            int([nIndividuals X 1]) 
  
  Todo: 
    * Extend to N objectives
  """
  fronts = getFronts(objVals)

  # Rank each individual in each front by crowding distance
  for f in range(len(fronts)):
    x1 = objVals[fronts[f],0]
    x2 = objVals[fronts[f],1]    
    crowdDist = getCrowdingDist(x1) + getCrowdingDist(x2)
    frontRank = np.argsort(-crowdDist)
    fronts[f] = [fronts[f][i] for i in frontRank]
    
  # Convert to ranking
  tmp = [ind for front in fronts for ind in front]  
  rank = np.empty_like(tmp)
  rank[tmp] = np.arange(len(tmp))

  if returnFronts is True:
    return rank, fronts
  else:
    return rank

def getFronts(objVals):
  """Fast non-dominated sort.
  
  Args: 
    objVals - (np_array) - Objective values of each individual
              [nInds X nObjectives]
      
  Returns: 
    front   - [list of lists] - One list for each front: 
                                list of indices of individuals in front
    
  Todo: 
    * Extend to N objectives

  [adapted from: https://github.com/haris989/NSGA-II]
  """
    
  values1 = objVals[:,0]
  values2 = objVals[:,1]
  
  S=[[] for i in range(0,len(values1))]
  front = [[]]
  n=[0 for i in range(0,len(values1))]
  rank = [0 for i in range(0, len(values1))]
  # Get domination relations
  for p in range(0,len(values1)):
      S[p]=[]
      n[p]=0
      for q in range(0, len(values1)):
          if (values1[p] > values1[q] and values2[p] > values2[q]) \
          or (values1[p] >= values1[q] and values2[p] > values2[q]) \
          or (values1[p] > values1[q] and values2[p] >= values2[q]):
              if q not in S[p]:
                  S[p].append(q)
          elif (values1[q] > values1[p] and values2[q] > values2[p]) \
          or (values1[q] >= values1[p] and values2[q] > values2[p]) \
          or (values1[q] > values1[p] and values2[q] >= values2[p]):
              n[p] = n[p] + 1
      if n[p]==0:
          rank[p] = 0
          if p not in front[0]:
              front[0].append(p)

  # Assign fronts
  i = 0
  while(front[i] != []):
      Q=[]
      for p in front[i]:
          for q in S[p]:
              n[q] =n[q] - 1
              if( n[q]==0):
                  rank[q]=i+1
                  if q not in Q:
                      Q.append(q)
      i = i+1
      front.append(Q)
  del front[len(front)-1]
  return front

def getCrowdingDist(objVector):
  """Returns crowding distance of a vector of values, used once on each front.

  Note: Crowding distance of individuals at each end of front is infinite, as they don't have a neighbor.

  Args: 
    objVector - (np_array) - Objective values of each individual
                [nInds X nObjectives]      
      
  Returns: 
    dist      - (np_array) - Crowding distance of each individual
                [nIndividuals X 1]
  """
  # Order by objective value
  key = np.argsort(objVector)
  sortedObj = objVector[key]
    
  # Distance from values on either side
  shiftVec = np.r_[np.inf,sortedObj,np.inf] # Edges have infinite distance
  warnings.filterwarnings("ignore", category=RuntimeWarning) # inf on purpose
  prevDist = np.abs(sortedObj-shiftVec[:-2])
  nextDist = np.abs(sortedObj-shiftVec[2:])
  crowd = prevDist+nextDist
  if (sortedObj[-1]-sortedObj[0]) > 0:
    crowd *= abs((1/sortedObj[-1]-sortedObj[0])) # Normalize by fitness range
  
  # Restore original order
  dist = np.empty(len(key))
  dist[key] = crowd[:]

  return dist


"""
Laconic plot functions to replace some of the matplotlibs verbosity
"""
from matplotlib import pyplot as plt
import numpy as np
import seaborn as sns


# -- File I/O ------------------------------------------------------------ -- #
def lsave(data,fileName):
  np.savetxt(fileName, data, delimiter=',',fmt='%1.2e')    
    
def lload(fileName):
  return np.loadtxt(fileName, delimiter=',') 


# -- Basic Plotting ------------------------------------------------------ -- #
def lplot(*args,label=[],axis=False):
  """Plots an vector, a set of vectors, with or without an x scale
  """
  fig, ax = getAxis(axis)

  if len(args) == 1: # No xscale
    x = np.arange(np.shape(args)[1])
    y = args[0]
  if len(args) == 2: # xscale given
    x = args[0]
    y = args[1]
    
  if np.ndim(y) == 2:
    for i in range(np.shape(y)[1]):
      ax.plot(x,y[:,i],'-')
      if len(label) > 0:
        ax.legend((label))     
  else:
    ax.plot(x,y,'o-')    
    
  if axis is False:    
    return fig, ax
  else:
    return ax

def ldist(x, axis=False):
  """Plots histogram with estimated distribution
  """
  fig, ax = getAxis(axis)
  
  if isinstance(x, str):
    vals = lload(x)
  else:
    vals = x
  sns.distplot(vals.flatten(),ax=ax,bins=10)
  #sns.distplot(vals.flatten(),ax=ax,hist_kws={"histtype": "step", "linewidth": 3, "alpha": 1, "color": "g"})
  return ax


def lquart(x,y,label=[],axis=False):
  """Plots quartiles, x is a vector, y is a matrix with same length as x
  """
  if axis is not False:
    ax = axis
    fig = ax.figure.canvas 
  else:
    fig, ax = plt.subplots()
  
  q = np.percentile(y,[25,50,75],axis=1)
  plt.plot(x,q[1,:],label=label) # median
  plt.plot(x,q[0,:],'k:',alpha=0.5)
  plt.plot(x,q[2,:],'k:',alpha=0.5)
  plt.fill_between(x,q[0,:],q[2,:],alpha=0.25)
  
  return ax
  
def getAxis(axis):
  if axis is not False:
    ax = axis
    fig = ax.figure.canvas 
  else:
    fig, ax = plt.subplots()
    
  return fig,ax
# -- --------------- -- --------------------------------------------#
  
  



import numpy as np
import math
import copy
import json

from .utils import *
from .nsga_sort import nsga_sort
from .ind import Ind


class Neat():
  """NEAT main class. Evolves population given fitness values of individuals.
  """
  def __init__(self, hyp):
    """Intialize NEAT algorithm with hyperparameters
    Args:
      hyp - (dict) - algorithm hyperparameters

    Attributes:
      p       - (dict)     - algorithm hyperparameters (see config/hypkey.txt)
      pop     - (Ind)      - Current population
      species - (Species)  - Current species   
      innov   - (np_array) - innovation record
                [5 X nUniqueGenes]
                [0,:] == Innovation Number
                [1,:] == Source
                [2,:] == Destination
                [3,:] == New Node?
                [4,:] == Generation evolved
      gen     - (int)      - Current generation
    """
    self.p       = hyp
    self.pop     = [] 
    self.species = [] 
    self.innov   = [] 
    self.gen     = 0  

  ''' Subfunctions '''
  from ._variation import evolvePop, recombine
  from ._speciate  import Species, speciate, compatDist,\
                          assignSpecies, assignOffspring  

  def ask(self):
    """Returns newly evolved population
    """
    if len(self.pop) == 0:
      self.initPop()      # Initialize population
    else:
      self.probMoo()      # Rank population according to objectivess
      self.speciate()     # Divide population into species
      self.evolvePop()    # Create child population 

    return self.pop       # Send child population for evaluation

  def tell(self,reward):
    """Assigns fitness to current population

    Args:
      reward - (np_array) - fitness value of each individual
               [nInd X 1]

    """
    for i in range(np.shape(reward)[0]):
      self.pop[i].fitness = reward[i]
      self.pop[i].nConn   = self.pop[i].nConn
  
  def initPop(self):
    """Initialize population with a list of random individuals
    """
    ##  Create base individual
    p = self.p # readability
    
    # - Create Nodes -
    nodeId = np.arange(0,p['ann_nInput']+ p['ann_nOutput']+1,1)
    node = np.empty((3,len(nodeId)))
    node[0,:] = nodeId
    
    # Node types: [1:input, 2:hidden, 3:bias, 4:output]
    node[1,0]             = 4 # Bias
    node[1,1:p['ann_nInput']+1] = 1 # Input Nodes
    node[1,(p['ann_nInput']+1):\
           (p['ann_nInput']+p['ann_nOutput']+1)]  = 2 # Output Nodes
    
    # Node Activations
    node[2,:] = p['ann_initAct']
    # - Create Conns -
    nConn = (p['ann_nInput']+1) * p['ann_nOutput']
    ins   = np.arange(0,p['ann_nInput']+1,1)            # Input and Bias Ids
    outs  = (p['ann_nInput']+1) + np.arange(0,p['ann_nOutput']) # Output Ids
    
    conn = np.empty((5,nConn,))
    conn[0,:] = np.arange(0,nConn,1)      # Connection Id
    conn[1,:] = np.tile(ins, len(outs))   # Source Nodes
    conn[2,:] = np.repeat(outs,len(ins) ) # Destination Nodes
    conn[3,:] = np.nan                    # Weight Values
    conn[4,:] = 1                         # Enabled?
        
    # Create population of individuals with varied weights
    pop = []
    for i in range(p['popSize']):
        newInd = Ind(conn, node)
        newInd.conn[3,:] = (2*(np.random.rand(1,nConn)-0.5))*p['ann_absWCap']
        newInd.conn[4,:] = np.random.rand(1,nConn) < p['prob_initEnable']
        newInd.express()
        newInd.birth = 0
        pop.append(copy.deepcopy(newInd))  
    # - Create Innovation Record -
    innov = np.zeros([5,nConn])
    innov[0:3,:] = pop[0].conn[0:3,:]
    innov[3,:] = -1
    
    self.pop = pop
    self.innov = innov

  def probMoo(self):
    """Rank population according to Pareto dominance.
    """
    # Compile objectives
    meanFit = np.asarray([ind.fitness for ind in self.pop])
    nConns  = np.asarray([ind.nConn   for ind in self.pop])
    nConns[nConns==0] = 1 # No connections is pareto optimal but boring...
    objVals = np.c_[meanFit,1/nConns] # Maximize

    # Alternate between two objectives and single objective
    if self.p['alg_probMoo'] < np.random.rand():
      rank = nsga_sort(objVals[:,[0,1]])
    else: # Single objective
      rank = rankArray(-objVals[:,0])

    # Assign ranks
    for i in range(len(self.pop)):
      self.pop[i].rank = rank[i]



import numpy as np

def roulette(pArr):
  """Returns random index, with each choices chance weighted
  Args:
    pArr    - (np_array) - vector containing weighting of each choice
              [N X 1]

  Returns:
    choice  - (int)      - chosen index
  """
  spin = np.random.rand()*np.sum(pArr)
  slot = pArr[0]
  choice = len(pArr)
  for i in range(1,len(pArr)):
    if spin < slot:
      choice = i
      break
    else:
      slot += pArr[i]
  return choice

def listXor(b,c):
  """Returns elements in lists b and c they don't share
  """
  A = [a for a in b+c if (a not in b) or (a not in c)]
  return A

def rankArray(X):
  """Returns ranking of a list, with ties resolved by first-found first-order
  NOTE: Sorts descending to follow numpy conventions
  """ 
  tmp = np.argsort(X)
  rank = np.empty_like(tmp)
  rank[tmp] = np.arange(len(X))
  return rank

def tiedRank(X):  
  """Returns ranking of a list, with ties recieving and averaged rank
  # Modified from: github.com/cmoscardi/ox_ml_practical/blob/master/util.py
  """
  Z = [(x, i) for i, x in enumerate(X)]  
  Z.sort(reverse=True)  
  n = len(Z)  
  Rx = [0]*n   
  start = 0 # starting mark  
  for i in range(1, n):  
     if Z[i][0] != Z[i-1][0]:
       for j in range(start, i):  
         Rx[Z[j][1]] = float(start+1+i)/2.0;
       start = i
  for j in range(start, n):  
    Rx[Z[j][1]] = float(start+1+n)/2.0;

  return np.asarray(Rx)

def bestIntSplit(ratio, total):
  """Divides a total into integer shares that best reflects ratio
    Args:
      share      - [1 X N ] - Percentage in each pile
      total      - [int   ] - Integer total to split
    
    Returns:
      intSplit   - [1 x N ] - Number in each pile
  """
  # Handle poorly defined ratio
  if sum(ratio) != 1:
    ratio = np.asarray(ratio)/sum(ratio)
  
  # Get share in real and integer values
  floatSplit = np.multiply(ratio,total)
  intSplit   = np.floor(floatSplit)
  remainder  = int(total - sum(intSplit))
  
  # Rank piles by most cheated by rounding
  deserving = np.argsort(-(floatSplit-intSplit),axis=0)
  
  # Distribute remained to most deserving
  intSplit[deserving[:remainder]] = intSplit[deserving[:remainder]] + 1    
  return intSplit

def quickINTersect(A,B):
  """ Faster set intersect: only valid for vectors of positive integers.
  (useful for matching indices)
    
    Example:
    A = np.array([0,1,2,3,5],dtype=np.int16)
    B = np.array([0,1,6,5],dtype=np.int16)
    C = np.array([0],dtype=np.int16)
    D = np.array([],dtype=np.int16)

    print(quickINTersect(A,B))
    print(quickINTersect(B,C))
    print(quickINTersect(B,D))
  """
  if (len(A) == 0) or (len(B) == 0):
    return [],[]
  P = np.zeros((1+max(max(A),max(B))),dtype=bool)
  P[A] = True
  IB = P[B]
  P[A] = False # Reset
  P[B] = True
  IA = P[A]

  return IA, IB



import numpy as np
import gym

from scipy.misc import imresize as resize
from gym.spaces.box import Box
from gym.envs.box2d.car_racing import CarRacing

from vae.vae import ConvVAE

from config import games

import json

#import imageio

SCREEN_X = 64
SCREEN_Y = 64

TIME_LIMIT = 1000

MU_MODE = True

def _clip(x, lo=0.0, hi=1.0):
  return np.minimum(np.maximum(x, lo), hi)

def _process_frame(frame):
  obs = frame[0:84, :, :].astype(np.float)/255.0
  obs = resize(obs, (64, 64))
  obs = ((1.0 - obs) * 255).round().astype(np.uint8)
  return obs

def _process_frame_green(frame):
  obs = frame[0:84, :, :].astype(np.float)/255.0
  obs = resize(obs, (64, 64))
  obs = ((1.0 - obs) * 255).round().astype(np.uint8)
  return obs[:, :, 1] # green channel

class VAERacing(CarRacing):
  def __init__(self, full_episode=False):
    super(VAERacing, self).__init__()
    self._internal_counter = 0
    self.z_size = games['vae_racing'].input_size
    self.vae = ConvVAE(batch_size=1, z_size=self.z_size, gpu_mode=False, is_training=False, reuse=True)
    self.vae.load_json('vae/vae_'+str(self.z_size)+'.json')
    self.full_episode = full_episode
    high = np.array([np.inf] * self.z_size)
    self.observation_space = Box(-high, high)
    self._has_rendered = False
    self.real_frame = None

  def reset(self):
    self._internal_counter = 0
    self._has_rendered = False
    self.real_frame = None
    return super(VAERacing, self).reset()

  def render(self, mode='human', close=False):
    if mode == 'human' or mode == 'rgb_array':
      self._has_rendered = True
    return super(VAERacing, self).render(mode=mode)

  def step(self, action):

    if not self._has_rendered:
      self.render("rgb_array")
      self._has_rendered = False

    if action is not None:
      action[0] = _clip(action[0], lo=-1.0, hi=+1.0)
      action[1] = _clip(action[1], lo=-1.0, hi=+1.0)
      action[1] = (action[1]+1.0) / 2.0
      action[2] = _clip(action[2])

    obs, reward, done, _ = super(VAERacing, self).step(action)

    result = np.copy(_process_frame(obs)).astype(np.float)/255.0
    result = result.reshape(1, 64, 64, 3)
    self.real_frame = result

    #z = self.vae.encode(result).flatten()
    mu, logvar = self.vae.encode_mu_logvar(result)
    mu = mu[0]
    logvar = logvar[0]
    s = logvar.shape
    z = mu + np.exp(logvar/2.0) * np.random.randn(*s)

    if self.full_episode:
      if MU_MODE:
        return mu, reward, False, {}
      else:
        return z, reward, False, {}

    self._internal_counter += 1
    if self._internal_counter > TIME_LIMIT:
      done = True

    if MU_MODE:
      return mu, reward, done, {}
    return z, reward, done, {}



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="prettyNEAT",
    version="0.0.2",
    author="Dennis G. Wilson",
    author_email="d9w@pm.me",
    description="NeuroEvolution of Augmenting Topologies in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/d9w/prettyNEAT",
    project_urls={
        "Bug Tracker": "https://github.com/d9w/prettyNEAT/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Artificial Life",
    ],
    packages=setuptools.find_packages(),
    install_requires=[
        "numpy",
    ],
    python_requires=">=3.5",
)



import numpy as np
from matplotlib.pyplot import imread


def make_env(env_name, seed=-1, render_mode=False):
  # -- Bullet Environments ------------------------------------------- -- #
  if "Bullet" in env_name:
    import pybullet as p # pip install pybullet
    import pybullet_envs
    import pybullet_envs.bullet.kukaGymEnv as kukaGymEnv

  # -- Bipedal Walker ------------------------------------------------ -- #
  if (env_name.startswith("BipedalWalker")):
    if (env_name.startswith("BipedalWalkerHardcore")):
      import Box2D
      from domain.bipedal_walker import BipedalWalkerHardcore
      env = BipedalWalkerHardcore()
    elif (env_name.startswith("BipedalWalkerMedium")): 
      from domain.bipedal_walker import BipedalWalker
      env = BipedalWalker()
      env.accel = 3
    else:
      from domain.bipedal_walker import BipedalWalker
      env = BipedalWalker()


  # -- VAE Racing ---------------------------------------------------- -- #
  elif (env_name.startswith("VAERacing")):
    from domain.vae_racing import VAERacing
    env = VAERacing()
    
    
  # -- Classification ------------------------------------------------ -- #
  elif (env_name.startswith("Classify")):
    from domain.classify_gym import ClassifyEnv
    if env_name.endswith("digits"):
      from domain.classify_gym import digit_raw
      trainSet, target  = digit_raw()
        
    if env_name.endswith("mnist256"):
      from domain.classify_gym import mnist_256
      trainSet, target  = mnist_256()

    env = ClassifyEnv(trainSet,target)  


  # -- Cart Pole Swing up -------------------------------------------- -- #
  elif (env_name.startswith("CartPoleSwingUp")):
    from domain.cartpole_swingup import CartPoleSwingUpEnv
    env = CartPoleSwingUpEnv()
    if (env_name.startswith("CartPoleSwingUp_Hard")):
      env.dt = 0.01
      env.t_limit = 200

  # -- Other  -------------------------------------------------------- -- #
  else:
    import gym
    env = gym.make(env_name)

  if (seed >= 0):
    domain.seed(seed)

  return env



import random
import numpy as np
import sys
from .make_env import make_env
from prettyNEAT import *

class GymTask():
  """Problem domain to be solved by neural network. Uses OpenAI Gym patterns.
  """ 
  def __init__(self, game, paramOnly=False, nReps=1): 
    """Initializes task environment
  
    Args:
      game - (string) - dict key of task to be solved (see domain/config.py)
  
    Optional:
      paramOnly - (bool)  - only load parameters instead of launching task?
      nReps     - (nReps) - number of trials to get average fitness
    """
    # Network properties
    self.nInput   = game.input_size
    self.nOutput  = game.output_size      
    self.actRange = game.h_act
    self.absWCap  = game.weightCap
    self.layers   = game.layers      
    self.activations = np.r_[np.full(1,1),game.i_act,game.o_act]
  
    # Environment
    self.nReps = nReps
    self.maxEpisodeLength = game.max_episode_length
    self.actSelect = game.actionSelect
    if not paramOnly:
      self.env = make_env(game.env_name)
    
    # Special needs...
    self.needsClosed = (game.env_name.startswith("CartPoleSwingUp"))    
  
  def getFitness(self, wVec, aVec, hyp=None, view=False, nRep=False, seed=-1):
    """Get fitness of a single individual.
  
    Args:
      wVec    - (np_array) - weight matrix as a flattened vector
                [N**2 X 1]
      aVec    - (np_array) - activation function of each node 
                [N X 1]    - stored as ints (see applyAct in ann.py)
  
    Optional:
      view    - (bool)     - view trial?
      nReps   - (nReps)    - number of trials to get average fitness
      seed    - (int)      - starting random seed for trials
  
    Returns:
      fitness - (float)    - mean reward over all trials
    """
    if nRep is False:
      nRep = self.nReps
    wVec[np.isnan(wVec)] = 0
    reward = np.empty(nRep)
    for iRep in range(nRep):
      reward[iRep] = self.testInd(wVec, aVec, view=view, seed=seed+iRep)
    fitness = np.mean(reward)
    return fitness

  def testInd(self, wVec, aVec, view=False,seed=-1):
    """Evaluate individual on task
    Args:
      wVec    - (np_array) - weight matrix as a flattened vector
                [N**2 X 1]
      aVec    - (np_array) - activation function of each node 
                [N X 1]    - stored as ints (see applyAct in ann.py)
  
    Optional:
      view    - (bool)     - view trial?
      seed    - (int)      - starting random seed for trials
  
    Returns:
      fitness - (float)    - reward earned in trial
    """
    if seed >= 0:
      random.seed(seed)
      np.random.seed(seed)
      self.env.seed(seed)
    state = self.env.reset()
    self.env.t = 0
    annOut = act(wVec, aVec, self.nInput, self.nOutput, state)  
    action = selectAct(annOut,self.actSelect)    
   
    wVec[wVec!=0]
    predName = str(np.mean(wVec[wVec!=0]))
    state, reward, done, info = self.env.step(action)
    
    if self.maxEpisodeLength == 0:
      if view:
        if self.needsClosed:
          self.env.render(close=done)  
        else:
          self.env.render()
      return reward
    else:
      totalReward = reward
    
    for tStep in range(self.maxEpisodeLength): 
      annOut = act(wVec, aVec, self.nInput, self.nOutput, state) 
      action = selectAct(annOut,self.actSelect) 
      state, reward, done, info = self.env.step(action)
      totalReward += reward  
      if view:
        if self.needsClosed:
          self.env.render(close=done)  
        else:
          self.env.render()
      if done:
        break
    return totalReward



import numpy as np
import copy
from .ann import getLayer, getNodeOrder


class Ind():
  """Individual class: genes, network, and fitness
  """ 
  def __init__(self, conn, node):
    """Intialize individual with given genes
    Args:
      conn - [5 X nUniqueGenes]
             [0,:] == Innovation Number
             [1,:] == Source
             [2,:] == Destination
             [3,:] == Weight
             [4,:] == Enabled?
      node - [3 X nUniqueGenes]
             [0,:] == Node Id
             [1,:] == Type (1=input, 2=output 3=hidden 4=bias)
             [2,:] == Activation function (as int)
  
    Attributes:
      node    - (np_array) - node genes (see args)
      conn    - (np_array) - conn genes (see args)
      nInput  - (int)      - number of inputs
      nOutput - (int)      - number of outputs
      wMat    - (np_array) - weight matrix, one row and column for each node
                [N X N]    - rows: connection from; cols: connection to
      wVec    - (np_array) - wMat as a flattened vector
                [N**2 X 1]    
      aVec    - (np_array) - activation function of each node (as int)
                [N X 1]    
      nConn   - (int)      - number of connections
      fitness - (double)   - fitness averaged over all trials (higher better)
      X fitMax  - (double)   - best fitness over all trials (higher better)
      rank    - (int)      - rank in population (lower better)
      birth   - (int)      - generation born
      species - (int)      - ID of species
    """
    self.node    = np.copy(node)
    self.conn    = np.copy(conn)
    self.nInput  = sum(node[1,:]==1)
    self.nOutput = sum(node[1,:]==2)
    self.wMat    = []
    self.wVec    = []
    self.aVec    = []
    self.nConn   = []
    self.fitness = [] # Mean fitness over trials
    #self.fitMax  = [] # Best fitness over trials
    self.rank    = []
    self.birth   = []
    self.species = []

  def nConns(self):
    """Returns number of active connections
    """
    return int(np.sum(self.conn[4,:]))

  def express(self):
    """Converts genes to weight matrix and activation vector
    """
    order, wMat = getNodeOrder(self.node, self.conn)
    if order is not False:
      self.wMat = wMat
      self.aVec = self.node[2,order]

      wVec = self.wMat.flatten()
      wVec[np.isnan(wVec)] = 0
      self.wVec  = wVec
      self.nConn = np.sum(wVec!=0)
      return True
    else:
      return False

  def createChild(self, p, innov, gen=0, mate=None):
    """Create new individual with this individual as a parent

      Args:
        p      - (dict)     - algorithm hyperparameters (see config/hypkey.txt)
        innov  - (np_array) - innovation record
           [5 X nUniqueGenes]
           [0,:] == Innovation Number
           [1,:] == Source
           [2,:] == Destination
           [3,:] == New Node?
           [4,:] == Generation evolved
        gen    - (int)      - (optional) generation (for innovation recording)
        mate   - (Ind)      - (optional) second for individual for crossover


    Returns:
        child  - (Ind)      - newly created individual
        innov  - (np_array) - updated innovation record

    """  
    if mate is not None:
      child = self.crossover(mate)
    else:
      child = Ind(self.conn, self.node)

    child, innov = child.mutate(p,innov,gen)
    return child, innov

# -- Canonical NEAT recombination operators ------------------------------ -- #

  def crossover(self,mate):
    """Combine genes of two individuals to produce new individual

      Procedure:
      ) Inherit all nodes and connections from most fit parent
      ) Identify matching connection genes in parentA and parentB
      ) Replace weights with parentB weights with some probability

      Args:
        parentA  - (Ind) - Fittest parent
          .conns - (np_array) - connection genes
                   [5 X nUniqueGenes]
                   [0,:] == Innovation Number (unique Id)
                   [1,:] == Source Node Id
                   [2,:] == Destination Node Id
                   [3,:] == Weight Value
                   [4,:] == Enabled?             
        parentB - (Ind) - Less fit parent

    Returns:
        child   - (Ind) - newly created individual

    """
    parentA = self
    parentB = mate

    # Inherit all nodes and connections from most fit parent
    child = Ind(parentA.conn, parentA.node)
    
    # Identify matching connection genes in ParentA and ParentB
    aConn = np.copy(parentA.conn[0,:])
    bConn = np.copy(parentB.conn[0,:])
    matching, IA, IB = np.intersect1d(aConn,bConn,return_indices=True)
    
    # Replace weights with parentB weights with some probability
    bProb = 0.5
    bGenes = np.random.rand(1,len(matching))<bProb
    child.conn[3,IA[bGenes[0]]] = parentB.conn[3,IB[bGenes[0]]]
    
    return child

  def mutate(self,p,innov=None,gen=None):
    """Randomly alter topology and weights of individual

    Args:
      p        - (dict)     - algorithm hyperparameters (see config/hypkey.txt)
      child    - (Ind) - individual to be mutated
        .conns - (np_array) - connection genes
                 [5 X nUniqueGenes] 
                 [0,:] == Innovation Number (unique Id)
                 [1,:] == Source Node Id
                 [2,:] == Destination Node Id
                 [3,:] == Weight Value
                 [4,:] == Enabled?  
        .nodes - (np_array) - node genes
                 [3 X nUniqueGenes]
                 [0,:] == Node Id
                 [1,:] == Type (1=input, 2=output 3=hidden 4=bias)
                 [2,:] == Activation function (as int)
      innov    - (np_array) - innovation record
                 [5 X nUniqueGenes]
                 [0,:] == Innovation Number
                 [1,:] == Source
                 [2,:] == Destination
                 [3,:] == New Node?
                 [4,:] == Generation evolved

    Returns:
        child   - (Ind)      - newly created individual
        innov   - (np_array) - innovation record

    """
    # Readability
    nConn = np.shape(self.conn)[1]
    connG = np.copy(self.conn)
    nodeG = np.copy(self.node)

    
    # - Re-enable connections
    disabled  = np.where(connG[4,:] == 0)[0]
    reenabled = np.random.rand(1,len(disabled)) < p['prob_enable']
    connG[4,disabled] = reenabled
         
    # - Weight mutation
    # [Canonical NEAT: 10% of weights are fully random...but seriously?]
    mutatedWeights = np.random.rand(1,nConn) < p['prob_mutConn'] # Choose weights to mutate
    weightChange = mutatedWeights * np.random.randn(1,nConn) * p['ann_mutSigma']
    connG[3,:] += weightChange[0]
    
    # Clamp weight strength [ Warning given for nan comparisons ]  
    connG[3, (connG[3,:] >  p['ann_absWCap'])] =  p['ann_absWCap']
    connG[3, (connG[3,:] < -p['ann_absWCap'])] = -p['ann_absWCap']
    
    if (np.random.rand() < p['prob_addNode']) and np.any(connG[4,:]==1):
      connG, nodeG, innov = self.mutAddNode(connG, nodeG, innov, gen, p)
    
    if (np.random.rand() < p['prob_addConn']):
      connG, innov = self.mutAddConn(connG, nodeG, innov, gen, p) 
    
    child = Ind(connG, nodeG)
    child.birth = gen

    return child, innov

  def mutAddNode(self, connG, nodeG, innov, gen, p):
    """Add new node to genome

    Args:
      connG    - (np_array) - connection genes
                 [5 X nUniqueGenes] 
                 [0,:] == Innovation Number (unique Id)
                 [1,:] == Source Node Id
                 [2,:] == Destination Node Id
                 [3,:] == Weight Value
                 [4,:] == Enabled?  
      nodeG    - (np_array) - node genes
                 [3 X nUniqueGenes]
                 [0,:] == Node Id
                 [1,:] == Type (1=input, 2=output 3=hidden 4=bias)
                 [2,:] == Activation function (as int)
      innov    - (np_array) - innovation record
                 [5 X nUniqueGenes]
                 [0,:] == Innovation Number
                 [1,:] == Source
                 [2,:] == Destination
                 [3,:] == New Node?
                 [4,:] == Generation evolved
      gen      - (int) - current generation
      p        - (dict)     - algorithm hyperparameters (see config/hypkey.txt)


    Returns:
      connG    - (np_array) - updated connection genes
      nodeG    - (np_array) - updated node genes
      innov    - (np_array) - updated innovation record

    """
    if innov is None:
      newNodeId = int(max(nodeG[0,:]+1))
      newConnId = connG[0,-1]+1    
    else:
      newNodeId = int(max(innov[2,:])+1) # next node id is a running counter
      newConnId = innov[0,-1]+1 
       
    # Choose connection to split
    connActive = np.where(connG[4,:] == 1)[0]
    if len(connActive) < 1:
      return connG, nodeG, innov # No active connections, nothing to split
    connSplit  = connActive[np.random.randint(len(connActive))]
    
    # Create new node
    newActivation = p['ann_actRange'][np.random.randint(len(p['ann_actRange']))]
    newNode = np.array([[newNodeId, 3, newActivation]]).T
    
    # Add connections to and from new node
    # -- Effort is taken to minimize disruption from node addition:
    # The 'weight to' the node is set to 1, the 'weight from' is set to the
    # original  weight. With a near linear activation function the change in performance should be minimal.

    connTo    = connG[:,connSplit].copy()
    connTo[0] = newConnId
    connTo[2] = newNodeId
    connTo[3] = 1 # weight set to 1
      
    connFrom    = connG[:,connSplit].copy()
    connFrom[0] = newConnId + 1
    connFrom[1] = newNodeId
    connFrom[3] = connG[3,connSplit] # weight set to previous weight value   
        
    newConns = np.vstack((connTo,connFrom)).T
        
    # Disable original connection
    connG[4,connSplit] = 0
        
    # Record innovations
    if innov is not None:
      newInnov = np.empty((5,2))
      newInnov[:,0] = np.hstack((connTo[0:3], newNodeId, gen))   
      newInnov[:,1] = np.hstack((connFrom[0:3], -1, gen)) 
      innov = np.hstack((innov,newInnov))
      
    # Add new structures to genome
    nodeG = np.hstack((nodeG,newNode))
    connG = np.hstack((connG,newConns))
    
    return connG, nodeG, innov

  def mutAddConn(self, connG, nodeG, innov, gen, p):
    """Add new connection to genome.
    To avoid creating recurrent connections all nodes are first sorted into
    layers, connections are then only created from nodes to nodes of the same or
    later layers.


    Todo: check for preexisting innovations to avoid duplicates in same gen

    Args:
      connG    - (np_array) - connection genes
                 [5 X nUniqueGenes] 
                 [0,:] == Innovation Number (unique Id)
                 [1,:] == Source Node Id
                 [2,:] == Destination Node Id
                 [3,:] == Weight Value
                 [4,:] == Enabled?  
      nodeG    - (np_array) - node genes
                 [3 X nUniqueGenes]
                 [0,:] == Node Id
                 [1,:] == Type (1=input, 2=output 3=hidden 4=bias)
                 [2,:] == Activation function (as int)
      innov    - (np_array) - innovation record
                 [5 X nUniqueGenes]
                 [0,:] == Innovation Number
                 [1,:] == Source
                 [2,:] == Destination
                 [3,:] == New Node?
                 [4,:] == Generation evolved
      gen      - (int)      - current generation
      p        - (dict)     - algorithm hyperparameters (see config/hypkey.txt)


    Returns:
      connG    - (np_array) - updated connection genes
      innov    - (np_array) - updated innovation record

    """
    if innov is None:
      newConnId = connG[0,-1]+1
    else:
      newConnId = innov[0,-1]+1 

    nIns = len(nodeG[0,nodeG[1,:] == 1]) + len(nodeG[0,nodeG[1,:] == 4])
    nOuts = len(nodeG[0,nodeG[1,:] == 2])
    order, wMat = getNodeOrder(nodeG, connG)   # Topological Sort of Network
    hMat = wMat[nIns:-nOuts,nIns:-nOuts]
    hLay = getLayer(hMat)+1

    # To avoid recurrent connections nodes are sorted into layers, and connections are only allowed from lower to higher layers
    if len(hLay) > 0:
      lastLayer = max(hLay)+1
    else:
      lastLayer = 1
    L = np.r_[np.zeros(nIns), hLay, np.full((nOuts),lastLayer) ]
    nodeKey = np.c_[nodeG[0,order], L] # Assign Layers

    sources = np.random.permutation(len(nodeKey))
    for src in sources:
      srcLayer = nodeKey[src,1]
      dest = np.where(nodeKey[:,1] > srcLayer)[0]
      
      # Finding already existing connections:
      #   ) take all connection genes with this source (connG[1,:])
      #   ) take the destination of those genes (connG[2,:])
      #   ) convert to nodeKey index (Gotta be a better numpy way...)   
      srcIndx = np.where(connG[1,:]==nodeKey[src,0])[0]
      exist = connG[2,srcIndx]
      existKey = []
      for iExist in exist:
        existKey.append(np.where(nodeKey[:,0]==iExist)[0])
      dest = np.setdiff1d(dest,existKey) # Remove existing connections
      
      # Add a random valid connection
      np.random.shuffle(dest)
      if len(dest)>0:  # (there is a valid connection)
        connNew = np.empty((5,1))
        connNew[0] = newConnId
        connNew[1] = nodeKey[src,0]
        connNew[2] = nodeKey[dest[0],0]
        connNew[3] = (np.random.rand()-0.5)*2*p['ann_absWCap']
        connNew[4] = 1
        connG = np.c_[connG,connNew]

        # Record innovation
        if innov is not None:
          newInnov = np.hstack((connNew[0:3].flatten(), -1, gen))
          innov = np.hstack((innov,newInnov[:,None]))
        break;

    return connG, innov
